import pytest
from speechcraft import text2voice, voice2voice, voice2embedding
from media_toolkit import AudioFile
import os


# Global directory paths for use in both pytest and direct execution
BASE_DIR = os.path.dirname(__file__)
INPUT_DIR = os.path.join(BASE_DIR, "test_files")
OUTPUT_DIR = os.path.join(BASE_DIR, "output")


@pytest.fixture
def sample_text():
    return "I love society [laughs]! [happy] What a day to make voice overs with artificial intelligence."


@pytest.fixture
def input_dir():
    return INPUT_DIR


@pytest.fixture
def output_dir():
    return OUTPUT_DIR


@pytest.fixture(autouse=True)
def setup_output_dir(output_dir):
    os.makedirs(output_dir, exist_ok=True)


def test_text2voice(sample_text, output_dir):
    audio_numpy, sample_rate = text2voice(sample_text)
    assert audio_numpy is not None

    audio = AudioFile().from_np_array(audio_numpy, sr=sample_rate)
    assert audio is not None
    output_path = f"{output_dir}/en_speaker_3_i_love_socaity.wav"

    audio.save(output_path)
    assert os.path.exists(output_path)


def test_voice2embedding(input_dir, output_dir):
    # Test voice cloning
    voice2embedding(audio_file=f"{input_dir}/voice_clone_test_voice_1.wav", voice_name="hermine").save_to_speaker_lib()

    # Test TTS with cloned voice
    tts_new_speaker, sample_rate = text2voice("Test text", voice="hermine")
    assert tts_new_speaker is not None

    audio_with_cloned_voice = AudioFile().from_np_array(tts_new_speaker, sr=sample_rate)
    assert audio_with_cloned_voice is not None
    output_path = f"{output_dir}/hermine_i_love_socaity.wav"
    audio_with_cloned_voice.save(output_path)
    assert os.path.exists(output_path)


def test_voice2voice(input_dir, output_dir):
    v2v_audio_np, sample_rate = voice2voice(audio_file=f"{input_dir}/voice_clone_test_voice_2.wav", voice_name="hermine")
    assert v2v_audio_np is not None

    v2v_audio = AudioFile().from_np_array(v2v_audio_np, sr=sample_rate)
    output_path = f"{output_dir}/potter_to_hermine.wav"
    v2v_audio.save(output_path)
    assert os.path.exists(output_path)


def test_with_media_file(input_dir, output_dir):
    """Test voice embedding creation using AudioFile object from bytes IO."""
    audio_file = AudioFile().from_file(f"{input_dir}/voice_clone_test_voice_1.wav")
    embed = voice2embedding(audio_file=audio_file.to_bytes_io(), voice_name="hermine").save_to_speaker_lib()
    assert embed is not None


if __name__ == "__main__":
    test_with_media_file()



import uuid
from typing import Union

from fastapi import HTTPException

from speechcraft import VoiceEmbedding

from speechcraft.settings import ALLOW_EMBEDDING_SAVE_ON_SERVER

try:
    from fast_task_api import FastTaskAPI, JobProgress, AudioFile, MediaFile
except ImportError:
    raise ImportError(
        "Please install the full version of speechcraft with pip install speechcraft[full]"
        " to use the server functionality."
    )

import speechcraft as t2v
from speechcraft.supp.model_downloader import download_all_models_init

from speechcraft.supp.utils import encode_path_safe

app = FastTaskAPI(
    title="SpeechCraft",
    summary="Create audio from text, clone voices and use them. Convert voice2voice. "
            "Generative text-to-audio Bark model.",
    version="0.0.15",
    contact={
        "name": "SocAIty",
        "url": "https://github.com/SocAIty/speechcraft",
    }
)


@app.task_endpoint(path="/text2voice")
def text2voice(
        job_progress: JobProgress,
        text: str,
        voice: Union[str, MediaFile] = "en_speaker_3",
        semantic_temp: float = 0.7,
        semantic_top_k: int = 50,
        semantic_top_p: float = 0.95,
        coarse_temp: float = 0.7,
        coarse_top_k: int = 50,
        coarse_top_p: float = 0.95,
        fine_temp: float = 0.5
):
    """
    :param text: the text to be converted
    :param voice: the name of the voice to be used. Uses the pretrained voices which are stored in models/speakers folder.
        It is also possible to provide a full path.
    :return: the audio file as bytes
    """
    # If voice is a MediaFile, it is gonna be voice cloning mode.
    if isinstance(voice, MediaFile):
        try:
            voice_name = getattr(voice, "file_name", "embedding")
            voice_name = f"{voice_name}_{uuid.uuid4()}"
            voice = VoiceEmbedding.load(voice.to_bytes_io(), speaker_name=voice_name)
        except Exception:
            raise HTTPException(status_code=400, detail="Invalid voice file. Please provide valid .npz generated with voice2voice")
    else:
        try:
            if isinstance(voice, str) and len(voice) == 0:
                print("No voice provided, using default voice")
                voice = "en_speaker_3"

            loaded_voice = VoiceEmbedding.load(voice)
            voice_name = loaded_voice.name
        except Exception:
            raise HTTPException(status_code=400, detail=f"Invalid voice: {voice}")

    # validate parameters
    # remove any illegal characters from text
    message = "Started text2voice." if isinstance(voice, str) else f"Text2voice using the voice embedding {voice_name}."
    job_progress.set_status(progress=0.01, message=message)

    generated_audio_file, sample_rate = t2v.text2voice(
        text=text,
        voice=voice,
        semantic_temp=semantic_temp,
        semantic_top_k=semantic_top_k,
        semantic_top_p=semantic_top_p,
        coarse_temp=coarse_temp,
        coarse_top_k=coarse_top_k,
        coarse_top_p=coarse_top_p,
        fine_temp=fine_temp,
        progress_update_func=job_progress.set_status
    )

    # make a recognizable filename
    filename = text[:15] if len(text) > 15 else text
    filename = encode_path_safe(filename)
    filename = f"{filename}_{voice_name}.wav"
    af = AudioFile(file_name=filename).from_np_array(np_array=generated_audio_file, sr=sample_rate, file_type="wav")
    return af


@app.task_endpoint("/voice2embedding")
def voice2embedding(
        job_progress: JobProgress,
        audio_file: AudioFile,
        voice_name: str = "new_speaker",
        save: bool = ALLOW_EMBEDDING_SAVE_ON_SERVER
):
    """
    :param audio_file: the audio file as bytes 5-20s is good length
    :param voice_name: how the new voice / embedding is named
    :param save: if the embedding should be saved in the voice dir for reusage.
        Note: depending on the server settings this might not be allowed
    :return: the voice embedding as bytes
    """
    # create embedding vector
    bytesio = audio_file.to_bytes_io()
    job_progress.set_status(progress=0.1, message=f"Started embedding creation {voice_name}.")
    embedding = t2v.voice2embedding(audio_file=bytesio, voice_name=voice_name)

    # write voice embedding to file
    if save and ALLOW_EMBEDDING_SAVE_ON_SERVER:
        job_progress.set_status(progress=0.99, message=f"Saving embedding {voice_name} to library.")
        embedding.save_to_speaker_lib()

    mf = MediaFile(file_name=f"{voice_name}.npz").from_bytesio(embedding.to_bytes_io(), copy=False)
    return mf


@app.task_endpoint("/voice2voice")
def voice2voice(
        job_progress: JobProgress,
        audio_file: AudioFile,
        voice_name: str | MediaFile = "en_speaker_3",
        temp: float = 0.7
):
    """
    :param audio_file: the audio file as bytes 5-20s is good length
    :param voice_name: the new of the voice to convert to; or the voice embedding. String or MediaFile.
    :param temp: generation temperature (1.0 more diverse, 0.0 more conservative)
    :return: the converted audio file as bytes
    """
    job_progress.set_status(progress=0.01, message=f"Started voice2voice {voice_name}.")

    # inference
    audio_array, sample_rate = t2v.voice2voice(
        audio_file.to_bytes_io(), voice_name,
        temp=temp, progress_update_func=job_progress.set_status)

    job_progress.set_status(progress=0.99, message=f"Converting to audio_file {voice_name}.")

    # convert to file
    af = AudioFile(file_name=f"voice2voice_{voice_name}.wav").from_np_array(
        np_array=audio_array,
        sr=sample_rate,
        file_type="wav"
    )

    return af


def start_server(port: int = 8009):
    # first time load and install models
    download_all_models_init()
    app.start(port=port)


# start the server on provided port
if __name__ == "__main__":
    start_server()



import os
import uuid
from io import BytesIO
from typing import Union, Optional

import numpy as np

from speechcraft.settings import EMBEDDINGS_DIR, DEFAULT_EMBEDDINGS_DIR


class VoiceEmbedding:
    def __init__(
            self,
            codes: np.array,
            semantic_tokens: np.array,
            name: str = "new_speaker"
    ):
        self.name = name
        self.fine_prompt = codes
        self.coarse_prompt = codes[:2, :]
        self.semantic_prompt = semantic_tokens

    def save_to_speaker_lib(self):
        sp = getattr(EMBEDDINGS_DIR, DEFAULT_EMBEDDINGS_DIR, "")
        self.save(sp)
        return self

    def save(self, folder: str):
        speaker_embedding_file = os.path.join(folder, f"{self.name}.npz")
        np.savez(speaker_embedding_file,
                 fine_prompt=self.fine_prompt,
                 coarse_prompt=self.coarse_prompt,
                 semantic_prompt=self.semantic_prompt
        )
        return speaker_embedding_file

    def to_bytes_io(self):
        f = BytesIO()
        np.savez(
           f, fine_prompt=self.fine_prompt, coarse_prompt=self.coarse_prompt,semantic_prompt=self.semantic_prompt
        )
        f.seek(0)
        return f

    @staticmethod
    def _load_embedding_data(source: Union[str, BytesIO]) -> tuple:
        """
        Internal method to load embedding data from a file or BytesIO object.

        Args:
            source: File path or BytesIO object containing the embedding data

        Returns:
            tuple: (codes, semantic_tokens)

        Raises:
            ValueError: If the embedding data is invalid or cannot be loaded
        """
        try:
            with np.load(source) as data:
                return data['fine_prompt'], data['semantic_prompt']
        except (KeyError, ValueError) as e:
            raise ValueError(f"Invalid embedding format. Error: {str(e)}")
        except Exception as e:
            raise ValueError(f"Failed to load embedding. Error: {str(e)}")

    @staticmethod
    def _find_embedding_file(path_or_name: str) -> Optional[str]:
        """
        Search for embedding file in multiple locations.

        Args:
            path_or_name: File path or speaker name

        Returns:
            str or None: Found file path or None if not found
        """
        # Try direct path first
        if os.path.isfile(path_or_name):
            return path_or_name

        filename = os.path.basename(path_or_name)
        search_paths = []

        # Add EMBEDDINGS_DIR to search paths if it exists and is valid
        if EMBEDDINGS_DIR and os.path.isdir(EMBEDDINGS_DIR):
            search_paths.append(EMBEDDINGS_DIR)

        # Always include DEFAULT_EMBEDDINGS_DIR
        search_paths.append(DEFAULT_EMBEDDINGS_DIR)

        # Search in all paths
        for directory in search_paths:
            filepath = os.path.join(directory, filename)
            if os.path.isfile(filepath):
                return filepath

        return None

    @classmethod
    def load(cls, source: Union[str, BytesIO], speaker_name: Optional[str] = None) -> "VoiceEmbedding":
        """
        Unified method to load a VoiceEmbedding from various sources.

        Args:
            source: Can be one of:
                - Full path to .npz file (absolute or relative)
                - Speaker name
                - BytesIO object containing embedding data
            speaker_name: Optional speaker name. If not provided:
                - For file paths: extracted from filename
                - For BytesIO: generated using UUID

        Returns:
            VoiceEmbedding: Loaded voice embedding object

        Raises:
            FileNotFoundError: If embedding file cannot be found
            ValueError: If embedding data has invalid format
        """
        if isinstance(source, BytesIO):
            # Handle BytesIO input
            codes, semantic_tokens = cls._load_embedding_data(source)
            final_speaker_name = speaker_name or f"embedding_{uuid.uuid4()}"
        else:
            # Handle string input (file path or speaker name)
            # Add .npz extension if not present
            source = f"{source}.npz" if not source.endswith('.npz') else source
            filepath = cls._find_embedding_file(source)

            if not filepath:
                searched_paths = [
                    source,
                    os.path.join(EMBEDDINGS_DIR, os.path.basename(source)) if EMBEDDINGS_DIR else None,
                    os.path.join(DEFAULT_EMBEDDINGS_DIR, os.path.basename(source))
                ]
                searched_paths = [p for p in searched_paths if p]  # Remove None values
                raise FileNotFoundError(
                    f"Voice embedding not found. Searched in:\n  - " +
                    "\n  - ".join(searched_paths)
                )

            codes, semantic_tokens = cls._load_embedding_data(filepath)
            final_speaker_name = speaker_name or os.path.splitext(os.path.basename(filepath))[0]

        return cls(
            name=final_speaker_name,
            codes=codes,
            semantic_tokens=semantic_tokens
        )

    def __getitem__(self, key):
        """
        Support of ["semantic_tokens"] and ["fine_prompt"] and ["coarse_prompt"] syntax.
        :param key: str
        """
        if hasattr(self, key):
            return getattr(self, key)
        raise KeyError(f"Key {key} not found in VoiceEmbedding")

    def __setitem__(self, key, value):
        """
        Support of ["semantic_tokens"] and ["fine_prompt"] and ["coarse_prompt"] syntax.
        :param key: str
        """
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            raise KeyError(f"Key {key} not found in VoiceEmbedding")



import contextlib
import gc
import re

from encodec import EncodecModel
import funcy
import numpy as np
from scipy.special import softmax
import torch
import torch.nn.functional as F
import tqdm
from transformers import BertTokenizer
from huggingface_hub import hf_hub_download

from .model import GPTConfig, GPT
from .model_fine import FineGPT, FineGPTConfig
from speechcraft.settings import MODELS_DIR, USE_SMALL_MODELS, GLOBAL_ENABLE_MPS, OFFLOAD_CPU
import os
import logging

from .voice_embedding import VoiceEmbedding

if (
        torch.cuda.is_available() and
        hasattr(torch.cuda, "amp") and
        hasattr(torch.cuda.amp, "autocast") and
        hasattr(torch.cuda, "is_bf16_supported") and
        torch.cuda.is_bf16_supported()
):
    autocast = funcy.partial(torch.cuda.amp.autocast, dtype=torch.bfloat16)
else:
    @contextlib.contextmanager
    def autocast():
        yield

# hold models in global scope to lazy load
global models
models = {}

global models_devices
models_devices = {}

CONTEXT_WINDOW_SIZE = 1024

SEMANTIC_RATE_HZ = 49.9
SEMANTIC_VOCAB_SIZE = 10_000

CODEBOOK_SIZE = 1024
N_COARSE_CODEBOOKS = 2
N_FINE_CODEBOOKS = 8
COARSE_RATE_HZ = 75

SAMPLE_RATE = 24_000

SUPPORTED_LANGS = [
    ("English", "en"),
    ("German", "de"),
    ("Spanish", "es"),
    ("French", "fr"),
    ("Hindi", "hi"),
    ("Italian", "it"),
    ("Japanese", "ja"),
    ("Korean", "ko"),
    ("Polish", "pl"),
    ("Portuguese", "pt"),
    ("Russian", "ru"),
    ("Turkish", "tr"),
    ("Chinese", "zh"),
]

ALLOWED_PROMPTS = {"announcer"}
for _, lang in SUPPORTED_LANGS:
    for prefix in ("", f"v2{os.path.sep}"):
        for n in range(10):
            ALLOWED_PROMPTS.add(f"{prefix}{lang}_speaker_{n}")

logger = logging.getLogger(__name__)

REMOTE_MODEL_PATHS = {
    "text_small": {
        "repo_id": "suno/bark",
        "file_name": "text.pt",
    },
    "coarse_small": {
        "repo_id": "suno/bark",
        "file_name": "coarse.pt",
    },
    "fine_small": {
        "repo_id": "suno/bark",
        "file_name": "fine.pt",
    },
    "text": {
        "repo_id": "suno/bark",
        "file_name": "text_2.pt",
    },
    "coarse": {
        "repo_id": "suno/bark",
        "file_name": "coarse_2.pt",
    },
    "fine": {
        "repo_id": "suno/bark",
        "file_name": "fine_2.pt",
    },
}

if not hasattr(torch.nn.functional, 'scaled_dot_product_attention') and torch.cuda.is_available():
    logger.warning(
        "torch version does not support flash attention. You will get faster" +
        " inference speed by upgrade torch to newest nightly version."
    )


def _grab_best_device(use_gpu=True):
    if torch.cuda.device_count() > 0 and use_gpu:
        device = "cuda"
    elif torch.backends.mps.is_available() and use_gpu and GLOBAL_ENABLE_MPS:
        device = "mps"
    else:
        device = "cpu"
    return device


def _get_ckpt_path(model_type, use_small: bool = False, path: str = None):
    model_key = f"{model_type}_small" if use_small or USE_SMALL_MODELS else model_type
    model_name = REMOTE_MODEL_PATHS[model_key]["file_name"]
    if path is None:
        path = MODELS_DIR
    return os.path.join(path, f"{model_name}")


class InferenceContext:
    def __init__(self, benchmark=False):
        # we can't expect inputs to be the same length, so disable benchmarking by default
        self._chosen_cudnn_benchmark = benchmark
        self._cudnn_benchmark = None

    def __enter__(self):
        self._cudnn_benchmark = torch.backends.cudnn.benchmark
        torch.backends.cudnn.benchmark = self._chosen_cudnn_benchmark

    def __exit__(self, exc_type, exc_value, exc_traceback):
        torch.backends.cudnn.benchmark = self._cudnn_benchmark


if torch.cuda.is_available():
    torch.backends.cuda.matmul.allow_tf32 = True
    torch.backends.cudnn.allow_tf32 = True


@contextlib.contextmanager
def _inference_mode():
    with InferenceContext(), torch.inference_mode(), torch.no_grad(), autocast():
        yield


def _clear_cuda_cache():
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.synchronize()


def clean_models(model_key=None):
    global models
    model_keys = [model_key] if model_key is not None else list(models.keys())
    for k in model_keys:
        if k in models:
            del models[k]
    _clear_cuda_cache()
    gc.collect()


def _load_model(ckpt_path, device, use_small=False, model_type="text"):
    if model_type == "text":
        ConfigClass = GPTConfig
        ModelClass = GPT
    elif model_type == "coarse":
        ConfigClass = GPTConfig
        ModelClass = GPT
    elif model_type == "fine":
        ConfigClass = FineGPTConfig
        ModelClass = FineGPT
    else:
        raise NotImplementedError()
    model_key = f"{model_type}_small" if use_small or USE_SMALL_MODELS else model_type
    model_info = REMOTE_MODEL_PATHS[model_key]
    if not os.path.exists(ckpt_path):
        logger.info(f"{model_type} model not found, downloading into `{MODELS_DIR}`.")
        hf_hub_download(repo_id=model_info["repo_id"], filename=model_info["file_name"], local_dir=MODELS_DIR)
    checkpoint = torch.load(ckpt_path, map_location=device, weights_only=False)
    # this is a hack
    model_args = checkpoint["model_args"]
    if "input_vocab_size" not in model_args:
        model_args["input_vocab_size"] = model_args["vocab_size"]
        model_args["output_vocab_size"] = model_args["vocab_size"]
        del model_args["vocab_size"]
    gptconf = ConfigClass(**checkpoint["model_args"])
    model = ModelClass(gptconf)
    state_dict = checkpoint["model"]
    # fixup checkpoint
    unwanted_prefix = "_orig_mod."
    for k, v in list(state_dict.items()):
        if k.startswith(unwanted_prefix):
            state_dict[k[len(unwanted_prefix):]] = state_dict.pop(k)
    extra_keys = set(state_dict.keys()) - set(model.state_dict().keys())
    extra_keys = set([k for k in extra_keys if not k.endswith(".attn.bias")])
    missing_keys = set(model.state_dict().keys()) - set(state_dict.keys())
    missing_keys = set([k for k in missing_keys if not k.endswith(".attn.bias")])
    if len(extra_keys) != 0:
        raise ValueError(f"extra keys found: {extra_keys}")
    if len(missing_keys) != 0:
        raise ValueError(f"missing keys: {missing_keys}")
    model.load_state_dict(state_dict, strict=False)
    n_params = model.get_num_params()
    val_loss = checkpoint["best_val_loss"].item()
    logger.info(f"model loaded: {round(n_params / 1e6, 1)}M params, {round(val_loss, 3)} loss")
    model.eval()
    model.to(device)
    del checkpoint, state_dict
    _clear_cuda_cache()
    if model_type == "text":
        tokenizer = BertTokenizer.from_pretrained("bert-base-multilingual-cased")
        return {
            "model": model,
            "tokenizer": tokenizer,
        }
    return model


def _load_codec_model(device):
    model = EncodecModel.encodec_model_24khz()
    model.set_target_bandwidth(6.0)
    model.eval()
    model.to(device)
    _clear_cuda_cache()
    return model


def load_model(use_gpu=True, use_small=False, force_reload=False, model_type="text", path=None):
    _load_model_f = funcy.partial(_load_model, model_type=model_type, use_small=use_small)
    if model_type not in ("text", "coarse", "fine"):
        raise NotImplementedError()
    global models
    global models_devices
    device = _grab_best_device(use_gpu=use_gpu)
    model_key = f"{model_type}"
    if OFFLOAD_CPU:
        models_devices[model_key] = device
        device = "cpu"
    if model_key not in models or force_reload:
        if path.endswith(".ckpt") or path.endswith(".pt") or path.endswith(".bin"):
            ckpt_path = path
        else:
            ckpt_path = _get_ckpt_path(model_type, use_small=use_small, path=path)
        # clean_models(model_key=model_key)
        model = _load_model_f(ckpt_path, device)
        models[model_key] = model
    if model_type == "text":
        models[model_key]["model"].to(device)
    else:
        models[model_key].to(device)
    return models[model_key]


def load_codec_model(use_gpu=True, force_reload=False):
    global models
    global models_devices
    device = _grab_best_device(use_gpu=use_gpu)
    if device == "mps":
        # encodec doesn't support mps
        device = "cpu"
    model_key = "codec"
    if OFFLOAD_CPU:
        models_devices[model_key] = device
        device = "cpu"
    if model_key not in models or force_reload:
        clean_models(model_key=model_key)
        model = _load_codec_model(device)
        models[model_key] = model
    models[model_key].to(device)
    return models[model_key]


def preload_models(
        text_use_gpu=True,
        text_use_small=False,
        text_model_path=None,
        coarse_use_gpu=True,
        coarse_use_small=False,
        coarse_model_path=None,
        fine_use_gpu=True,
        fine_use_small=False,
        fine_model_path=None,
        codec_use_gpu=True,
        force_reload=False,
        path=None,
):
    """Load all the necessary models for the pipeline."""
    if _grab_best_device() == "cpu" and (
            text_use_gpu or coarse_use_gpu or fine_use_gpu or codec_use_gpu
    ):
        logger.warning("No GPU being used. Careful, inference might be very slow!")

    _ = load_model(
        model_type="text", use_gpu=text_use_gpu, use_small=text_use_small, force_reload=force_reload,
        path=path if text_model_path is None else text_model_path
    )
    _ = load_model(
        model_type="coarse",
        use_gpu=coarse_use_gpu,
        use_small=coarse_use_small,
        force_reload=force_reload,
        path=path if coarse_model_path is None else coarse_model_path,
    )
    _ = load_model(
        model_type="fine", use_gpu=fine_use_gpu, use_small=fine_use_small, force_reload=force_reload,
        path=path if fine_model_path is None else fine_model_path
    )
    _ = load_codec_model(use_gpu=codec_use_gpu, force_reload=force_reload)


####
# Generation Functionality
####


def _tokenize(tokenizer, text):
    return tokenizer.encode(text, add_special_tokens=False)


def _detokenize(tokenizer, enc_text):
    return tokenizer.decode(enc_text)


def _normalize_whitespace(text):
    return re.sub(r"\s+", " ", text).strip()


TEXT_ENCODING_OFFSET = 10_048
SEMANTIC_PAD_TOKEN = 10_000
TEXT_PAD_TOKEN = 129_595
SEMANTIC_INFER_TOKEN = 129_599


def generate_text_semantic(
        text,
        history_prompt=None,
        temp=0.7,
        top_k=None,
        top_p=None,
        min_eos_p=0.2,
        max_gen_duration_s=None,
        allow_early_stop=True,
        use_kv_caching=False,
        progress_update_func=None
):
    """Generate semantic tokens from text."""
    assert isinstance(text, str)
    text = _normalize_whitespace(text)
    assert len(text.strip()) > 0
    if history_prompt is not None:
        if not isinstance(history_prompt, VoiceEmbedding):
            history_prompt = VoiceEmbedding.load(history_prompt)
        semantic_history = history_prompt["semantic_prompt"]
        assert (
                isinstance(semantic_history, np.ndarray)
                and len(semantic_history.shape) == 1
                and len(semantic_history) > 0
                and semantic_history.min() >= 0
                and semantic_history.max() <= SEMANTIC_VOCAB_SIZE - 1
        )
    else:
        semantic_history = None
    # load models if not yet exist
    global models
    global models_devices
    if "text" not in models:
        preload_models()
    model_container = models["text"]
    model = model_container["model"]
    tokenizer = model_container["tokenizer"]
    encoded_text = np.array(_tokenize(tokenizer, text)) + TEXT_ENCODING_OFFSET
    if OFFLOAD_CPU:
        model.to(models_devices["text"])
    device = next(model.parameters()).device
    if len(encoded_text) > 256:
        p = round((len(encoded_text) - 256) / len(encoded_text) * 100, 1)
        logger.warning(f"warning, text too long, lopping of last {p}%")
        encoded_text = encoded_text[:256]
    encoded_text = np.pad(
        encoded_text,
        (0, 256 - len(encoded_text)),
        constant_values=TEXT_PAD_TOKEN,
        mode="constant",
    )
    if semantic_history is not None:
        semantic_history = semantic_history.astype(np.int64)
        # lop off if history is too long, pad if needed
        semantic_history = semantic_history[-256:]
        semantic_history = np.pad(
            semantic_history,
            (0, 256 - len(semantic_history)),
            constant_values=SEMANTIC_PAD_TOKEN,
            mode="constant",
        )
    else:
        semantic_history = np.array([SEMANTIC_PAD_TOKEN] * 256)
    x = torch.from_numpy(
        np.hstack([
            encoded_text, semantic_history, np.array([SEMANTIC_INFER_TOKEN])
        ]).astype(np.int64)
    )[None]
    assert x.shape[1] == 256 + 256 + 1
    with _inference_mode():
        x = x.to(device)
        n_tot_steps = 768
        # if no progress_update_func, use tqdm
        # custom tqdm updates since we don't know when eos will occur
        progress_update_func = _get_progress_update_func(progress_update_func)

        pbar_state = 0
        tot_generated_duration_s = 0
        kv_cache = None
        for n in range(n_tot_steps):
            if use_kv_caching and kv_cache is not None:
                x_input = x[:, [-1]]
            else:
                x_input = x
            logits, kv_cache = model(
                x_input, merge_context=True, use_cache=use_kv_caching, past_kv=kv_cache
            )
            relevant_logits = logits[0, 0, :SEMANTIC_VOCAB_SIZE]
            if allow_early_stop:
                relevant_logits = torch.hstack(
                    (relevant_logits, logits[0, 0, [SEMANTIC_PAD_TOKEN]])  # eos
                )
            if top_p is not None:
                # faster to convert to numpy
                original_device = relevant_logits.device
                relevant_logits = relevant_logits.detach().cpu().type(torch.float32).numpy()
                sorted_indices = np.argsort(relevant_logits)[::-1]
                sorted_logits = relevant_logits[sorted_indices]
                cumulative_probs = np.cumsum(softmax(sorted_logits))
                sorted_indices_to_remove = cumulative_probs > top_p
                sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].copy()
                sorted_indices_to_remove[0] = False
                relevant_logits[sorted_indices[sorted_indices_to_remove]] = -np.inf
                relevant_logits = torch.from_numpy(relevant_logits)
                relevant_logits = relevant_logits.to(original_device)
            if top_k is not None:
                v, _ = torch.topk(relevant_logits, min(top_k, relevant_logits.size(-1)))
                relevant_logits[relevant_logits < v[-1]] = -float("Inf")
            probs = F.softmax(relevant_logits / temp, dim=-1)
            item_next = torch.multinomial(probs, num_samples=1).to(torch.int32)
            if allow_early_stop and (
                    item_next == SEMANTIC_VOCAB_SIZE
                    or (min_eos_p is not None and probs[-1] >= min_eos_p)
            ):
                # eos found, so break
                progress_update_func(1.0)
                break
            x = torch.cat((x, item_next[None]), dim=1)
            tot_generated_duration_s += 1 / SEMANTIC_RATE_HZ
            # hard cut if audio becomes too long (>max_gen_duration)
            if max_gen_duration_s is not None and tot_generated_duration_s > max_gen_duration_s:
                progress_update_func(1.0)
                break
            # process completed
            if n == n_tot_steps - 1:
                progress_update_func(1.0)
                break
            del logits, relevant_logits, probs, item_next
            # normal progress bar update
            if n > n_tot_steps:
                progress_update_func(1.0)
            else:
                progress_update_func(round(n / n_tot_steps, 2))

        out = x.detach().cpu().numpy().squeeze()[256 + 256 + 1:]
    if OFFLOAD_CPU:
        model.to("cpu")
    assert all(0 <= out) and all(out < SEMANTIC_VOCAB_SIZE)
    _clear_cuda_cache()
    return out


def _flatten_codebooks(arr, offset_size=CODEBOOK_SIZE):
    assert len(arr.shape) == 2
    arr = arr.copy()
    if offset_size is not None:
        for n in range(1, arr.shape[0]):
            arr[n, :] += offset_size * n
    flat_arr = arr.ravel("F")
    return flat_arr


COARSE_SEMANTIC_PAD_TOKEN = 12_048
COARSE_INFER_TOKEN = 12_050


def _get_progress_update_func(progress_update_func: callable = None):
    # replace progress_update_func with tqdm if not provided to show progress
    if progress_update_func is not None:
        return progress_update_func

    # return a tqdm object to display progress.
    p = tqdm.tqdm(total=100)
    # wrap tqdm to make compatible with any func
    def upn(n):
        p.n = round(n*100, 2)
        p.refresh()
    return upn


def generate_coarse(
        x_semantic,
        history_prompt=None,
        temp=0.7,
        top_k=None,
        top_p=None,
        max_coarse_history=630,  # min 60 (faster), max 630 (more context)
        sliding_window_len=60,
        use_kv_caching=False,
        progress_update_func=None  # function will be called with progress_update_func(x) x in [0, 1]
):
    """Generate coarse audio codes from semantic tokens."""
    assert (
            isinstance(x_semantic, np.ndarray)
            and len(x_semantic.shape) == 1
            and len(x_semantic) > 0
            and x_semantic.min() >= 0
            and x_semantic.max() <= SEMANTIC_VOCAB_SIZE - 1
    )
    assert 60 <= max_coarse_history <= 630
    assert max_coarse_history + sliding_window_len <= 1024 - 256
    semantic_to_coarse_ratio = COARSE_RATE_HZ / SEMANTIC_RATE_HZ * N_COARSE_CODEBOOKS
    max_semantic_history = int(np.floor(max_coarse_history / semantic_to_coarse_ratio))
    if history_prompt is not None:
        if not isinstance(history_prompt, VoiceEmbedding):
            history_prompt = VoiceEmbedding.load(history_prompt)
        x_semantic_history = history_prompt["semantic_prompt"]
        x_coarse_history = history_prompt["coarse_prompt"]
        assert (
                isinstance(x_semantic_history, np.ndarray)
                and len(x_semantic_history.shape) == 1
                and len(x_semantic_history) > 0
                and x_semantic_history.min() >= 0
                and x_semantic_history.max() <= SEMANTIC_VOCAB_SIZE - 1
                and isinstance(x_coarse_history, np.ndarray)
                and len(x_coarse_history.shape) == 2
                and x_coarse_history.shape[0] == N_COARSE_CODEBOOKS
                and x_coarse_history.shape[-1] >= 0
                and x_coarse_history.min() >= 0
                and x_coarse_history.max() <= CODEBOOK_SIZE - 1
                and (
                        round(x_coarse_history.shape[-1] / len(x_semantic_history), 1)
                        == round(semantic_to_coarse_ratio / N_COARSE_CODEBOOKS, 1)
                )
        )
        x_coarse_history = _flatten_codebooks(x_coarse_history) + SEMANTIC_VOCAB_SIZE
        # trim histories correctly
        n_semantic_hist_provided = np.min(
            [
                max_semantic_history,
                len(x_semantic_history) - len(x_semantic_history) % 2,
                int(np.floor(len(x_coarse_history) / semantic_to_coarse_ratio)),
            ]
        )
        n_coarse_hist_provided = int(round(n_semantic_hist_provided * semantic_to_coarse_ratio))
        x_semantic_history = x_semantic_history[-n_semantic_hist_provided:].astype(np.int32)
        x_coarse_history = x_coarse_history[-n_coarse_hist_provided:].astype(np.int32)
        # TODO: bit of a hack for time alignment (sounds better)
        x_coarse_history = x_coarse_history[:-2]
    else:
        x_semantic_history = np.array([], dtype=np.int32)
        x_coarse_history = np.array([], dtype=np.int32)
    # load models if not yet exist
    global models
    global models_devices
    if "coarse" not in models:
        preload_models()
    model = models["coarse"]
    if OFFLOAD_CPU:
        model.to(models_devices["coarse"])
    device = next(model.parameters()).device
    # start loop
    n_steps = int(
        round(
            np.floor(len(x_semantic) * semantic_to_coarse_ratio / N_COARSE_CODEBOOKS)
            * N_COARSE_CODEBOOKS
        )
    )
    assert n_steps > 0 and n_steps % N_COARSE_CODEBOOKS == 0
    x_semantic = np.hstack([x_semantic_history, x_semantic]).astype(np.int32)
    x_coarse = x_coarse_history.astype(np.int32)
    base_semantic_idx = len(x_semantic_history)
    with _inference_mode():
        x_semantic_in = torch.from_numpy(x_semantic)[None].to(device)
        x_coarse_in = torch.from_numpy(x_coarse)[None].to(device)
        n_window_steps = int(np.ceil(n_steps / sliding_window_len))
        n_step = 0

        # replace progress_update_func with tqdm if not provided to show progress
        progress_update_func = _get_progress_update_func(progress_update_func)

        for i in range(n_window_steps):
            # update progress bar
            progress_update_func(round(i / n_window_steps, 2))

            semantic_idx = base_semantic_idx + int(round(n_step / semantic_to_coarse_ratio))
            # pad from right side
            x_in = x_semantic_in[:, np.max([0, semantic_idx - max_semantic_history]):]
            x_in = x_in[:, :256]
            x_in = F.pad(
                x_in,
                (0, 256 - x_in.shape[-1]),
                "constant",
                COARSE_SEMANTIC_PAD_TOKEN,
            )
            x_in = torch.hstack(
                [
                    x_in,
                    torch.tensor([COARSE_INFER_TOKEN])[None].to(device),
                    x_coarse_in[:, -max_coarse_history:],
                ]
            )
            kv_cache = None
            for _ in range(sliding_window_len):
                if n_step >= n_steps:
                    continue
                is_major_step = n_step % N_COARSE_CODEBOOKS == 0

                if use_kv_caching and kv_cache is not None:
                    x_input = x_in[:, [-1]]
                else:
                    x_input = x_in

                logits, kv_cache = model(x_input, use_cache=use_kv_caching, past_kv=kv_cache)
                logit_start_idx = (
                        SEMANTIC_VOCAB_SIZE + (1 - int(is_major_step)) * CODEBOOK_SIZE
                )
                logit_end_idx = (
                        SEMANTIC_VOCAB_SIZE + (2 - int(is_major_step)) * CODEBOOK_SIZE
                )
                relevant_logits = logits[0, 0, logit_start_idx:logit_end_idx]
                if top_p is not None:
                    # faster to convert to numpy
                    original_device = relevant_logits.device
                    relevant_logits = relevant_logits.detach().cpu().type(torch.float32).numpy()
                    sorted_indices = np.argsort(relevant_logits)[::-1]
                    sorted_logits = relevant_logits[sorted_indices]
                    cumulative_probs = np.cumsum(softmax(sorted_logits))
                    sorted_indices_to_remove = cumulative_probs > top_p
                    sorted_indices_to_remove[1:] = sorted_indices_to_remove[:-1].copy()
                    sorted_indices_to_remove[0] = False
                    relevant_logits[sorted_indices[sorted_indices_to_remove]] = -np.inf
                    relevant_logits = torch.from_numpy(relevant_logits)
                    relevant_logits = relevant_logits.to(original_device)
                if top_k is not None:
                    v, _ = torch.topk(relevant_logits, min(top_k, relevant_logits.size(-1)))
                    relevant_logits[relevant_logits < v[-1]] = -float("Inf")
                probs = F.softmax(relevant_logits / temp, dim=-1)
                item_next = torch.multinomial(probs, num_samples=1).to(torch.int32)
                item_next += logit_start_idx
                x_coarse_in = torch.cat((x_coarse_in, item_next[None]), dim=1)
                x_in = torch.cat((x_in, item_next[None]), dim=1)
                del logits, relevant_logits, probs, item_next
                n_step += 1
            del x_in
        del x_semantic_in



    if OFFLOAD_CPU:
        model.to("cpu")
    gen_coarse_arr = x_coarse_in.detach().cpu().numpy().squeeze()[len(x_coarse_history):]
    del x_coarse_in
    assert len(gen_coarse_arr) == n_steps
    gen_coarse_audio_arr = gen_coarse_arr.reshape(-1, N_COARSE_CODEBOOKS).T - SEMANTIC_VOCAB_SIZE
    for n in range(1, N_COARSE_CODEBOOKS):
        gen_coarse_audio_arr[n, :] -= n * CODEBOOK_SIZE
    _clear_cuda_cache()
    return gen_coarse_audio_arr


def generate_fine(
        x_coarse_gen,
        history_prompt=None,
        temp=0.5,
        progress_update_func=None
):
    """Generate full audio codes from coarse audio codes."""
    assert (
            isinstance(x_coarse_gen, np.ndarray)
            and len(x_coarse_gen.shape) == 2
            and 1 <= x_coarse_gen.shape[0] <= N_FINE_CODEBOOKS - 1
            and x_coarse_gen.shape[1] > 0
            and x_coarse_gen.min() >= 0
            and x_coarse_gen.max() <= CODEBOOK_SIZE - 1
    )
    if history_prompt is not None:
        if not isinstance(history_prompt, VoiceEmbedding):
            history_prompt = VoiceEmbedding.load(history_prompt)
        x_fine_history = history_prompt["fine_prompt"]
        assert (
                isinstance(x_fine_history, np.ndarray)
                and len(x_fine_history.shape) == 2
                and x_fine_history.shape[0] == N_FINE_CODEBOOKS
                and x_fine_history.shape[1] >= 0
                and x_fine_history.min() >= 0
                and x_fine_history.max() <= CODEBOOK_SIZE - 1
        )
    else:
        x_fine_history = None
    n_coarse = x_coarse_gen.shape[0]
    # load models if not yet exist
    global models
    global models_devices
    if "fine" not in models:
        preload_models()
    model = models["fine"]
    if OFFLOAD_CPU:
        model.to(models_devices["fine"])
    device = next(model.parameters()).device
    # make input arr
    in_arr = np.vstack(
        [
            x_coarse_gen,
            np.zeros((N_FINE_CODEBOOKS - n_coarse, x_coarse_gen.shape[1]))
            + CODEBOOK_SIZE,  # padding
        ]
    ).astype(np.int32)
    # prepend history if available (max 512)
    if x_fine_history is not None:
        x_fine_history = x_fine_history.astype(np.int32)
        in_arr = np.hstack(
            [
                x_fine_history[:, -512:].astype(np.int32),
                in_arr,
            ]
        )
        n_history = x_fine_history[:, -512:].shape[1]
    else:
        n_history = 0
    n_remove_from_end = 0
    # need to pad if too short (since non-causal model)
    if in_arr.shape[1] < 1024:
        n_remove_from_end = 1024 - in_arr.shape[1]
        in_arr = np.hstack(
            [
                in_arr,
                np.zeros((N_FINE_CODEBOOKS, n_remove_from_end), dtype=np.int32) + CODEBOOK_SIZE,
            ]
        )
    # we can be lazy about fractional loop and just keep overwriting codebooks
    n_loops = np.max([0, int(np.ceil((x_coarse_gen.shape[1] - (1024 - n_history)) / 512))]) + 1
    with _inference_mode():
        in_arr = torch.tensor(in_arr.T).to(device)

        progress_update_func = _get_progress_update_func(progress_update_func)

        for n in range(n_loops):
            # update progress bar
            progress_update_func(round(n / n_loops, 2))

            start_idx = np.min([n * 512, in_arr.shape[0] - 1024])
            start_fill_idx = np.min([n_history + n * 512, in_arr.shape[0] - 512])
            rel_start_fill_idx = start_fill_idx - start_idx
            in_buffer = in_arr[start_idx: start_idx + 1024, :][None]
            for nn in range(n_coarse, N_FINE_CODEBOOKS):
                logits = model(nn, in_buffer)
                if temp is None:
                    relevant_logits = logits[0, rel_start_fill_idx:, :CODEBOOK_SIZE]
                    codebook_preds = torch.argmax(relevant_logits, -1)
                else:
                    relevant_logits = logits[0, :, :CODEBOOK_SIZE] / temp
                    probs = F.softmax(relevant_logits, dim=-1)
                    codebook_preds = torch.multinomial(
                        probs[rel_start_fill_idx:1024], num_samples=1
                    ).reshape(-1)
                codebook_preds = codebook_preds.to(torch.int32)
                in_buffer[0, rel_start_fill_idx:, nn] = codebook_preds
                del logits, codebook_preds
            # transfer over info into model_in and convert to numpy
            for nn in range(n_coarse, N_FINE_CODEBOOKS):
                in_arr[
                start_fill_idx: start_fill_idx + (1024 - rel_start_fill_idx), nn
                ] = in_buffer[0, rel_start_fill_idx:, nn]
            del in_buffer
        gen_fine_arr = in_arr.detach().cpu().numpy().squeeze().T
        del in_arr
    if OFFLOAD_CPU:
        model.to("cpu")
    gen_fine_arr = gen_fine_arr[:, n_history:]
    if n_remove_from_end > 0:
        gen_fine_arr = gen_fine_arr[:, :-n_remove_from_end]
    assert gen_fine_arr.shape[-1] == x_coarse_gen.shape[-1]
    _clear_cuda_cache()
    return gen_fine_arr


def codec_decode(fine_tokens):
    """Turn quantized audio codes into audio array using encodec."""
    # load models if not yet exist
    global models
    global models_devices
    if "codec" not in models:
        preload_models()
    model = models["codec"]
    if OFFLOAD_CPU:
        model.to(models_devices["codec"])
    device = next(model.parameters()).device
    arr = torch.from_numpy(fine_tokens)[None]
    arr = arr.to(device)
    arr = arr.transpose(0, 1)
    emb = model.quantizer.decode(arr)
    out = model.decoder(emb)
    audio_arr = out.detach().cpu().numpy().squeeze()
    del arr, emb, out
    if OFFLOAD_CPU:
        model.to("cpu")
    return audio_arr



from typing import Union
from io import BytesIO

import torchaudio
from encodec.utils import convert_audio
import numpy as np

import speechcraft.supp.utils as utils
from speechcraft.core.api import semantic_to_waveform
from speechcraft.settings import MODELS_DIR
from speechcraft.supp.model_downloader import get_hubert_manager_and_model, make_sure_models_are_downloaded


def voice2voice(
        audio_file: Union[BytesIO, str],
        voice_name: Union[BytesIO, str],
        temp: float = 0.7,
        max_coarse_history: int = 300,
        progress_update_func: callable = None
) -> tuple[np.ndarray, int]:
    """
    Takes voice and intonation from speaker_embedding and applies it to swap_audio_filename
    :param audio_file: the audio file to swap the voice. Can be a path or a file handle
    :param voice_name: the voice name or the voice embedding to use for the swap
    :param temp: generation temperature (1.0 more diverse, 0.0 more conservative)
    :param max_coarse_history: history influence. Min 60 (faster), max 630 (more context)
    :param progress_update_func: a callable to update the progress of the task.
        Called like progress_update_function(x) with x in [0, 1]
    :return:
    """
    #
    make_sure_models_are_downloaded(install_path=MODELS_DIR)
    # Load the HuBERT model
    hubert_manager, hubert_model, model, tokenizer = get_hubert_manager_and_model()

    # create a better progress function
    if progress_update_func is not None:
        progress_update_func = utils.create_progress_tracker(progress_update_func, steps=[
            ("loading", 10), ("embedding", 90)
        ])

    # Load and pre-process the audio waveform
    wav, sr = torchaudio.load(audio_file)
    if wav.shape[0] == 2:  # Stereo to mono if needed
        wav = wav.mean(0, keepdim=True)

    wav = convert_audio(wav, sr, model.sample_rate, model.channels)
    device = utils.get_cpu_or_gpu()
    wav = wav.to(device)

    if progress_update_func is not None:
        progress_update_func(50)  # 1 % for loading the audio

    # run inference
    print("embedding audio with hubert_model")
    semantic_vectors = hubert_model.forward(wav, input_sample_hz=model.sample_rate)
    semantic_tokens = tokenizer.get_token(semantic_vectors)

    # move semantic tokens to cpu
    semantic_tokens = semantic_tokens.cpu().numpy()

    if progress_update_func is not None:
        progress_update_func(100)  # Will start the next 'step' in the progress tracker

    # convert voice2voice
    print("inferencing")
    output_full = False
    out = semantic_to_waveform(
        semantic_tokens,
        history_prompt=voice_name,
        temp=temp,
        max_coarse_history=max_coarse_history,
        output_full=output_full,
        progress_update_func=progress_update_func
    )
    if output_full:
        full_generation, audio_arr = out
    else:
        audio_arr = out

    return audio_arr, model.sample_rate



import os

ROOT_DIR = os.getenv('ROOT_DIR', os.path.dirname(os.path.abspath(__file__)))
MODELS_DIR = os.getenv("MODELS_DIR", os.path.join(ROOT_DIR, "models"))

# If an embedding is loaded from file it will act in the following way:
# 1. Check if embedding can be found in the EMBEDDINGS_DIR
# 2. If not found, check if it can be found in the DEFAULT_EMBEDDINGS_DIR (is default speaker)
# For saving the same (if embeddings_dir is set save there, if not save in default)

DEFAULT_EMBEDDINGS_DIR = os.path.join(ROOT_DIR, "assets", "prompts")
EMBEDDINGS_DIR = os.getenv("EMBEDDINGS_DIR", None)
ALLOW_EMBEDDING_SAVE_ON_SERVER = os.getenv("ALLOW_EMBEDDING_SAVE_ON_SERVER", True) in ('true', '1', 't')

# Defining the getters and setters like this, allows a change of the env variable at runtime for other modules
# This is useful for the server, where the user can change the settings without restarting the server
# Required in cloud environments for example when a volume is mounted after the server is started
def set_embeddings_dir(path):
    global EMBEDDINGS_DIR
    EMBEDDINGS_DIR = path

def get_embeddings_dir():
    global EMBEDDINGS_DIR
    return EMBEDDINGS_DIR

# These settings are used in generation.py and are used in inference
def _cast_bool_env_var(s):
    return s.lower() in ('true', '1', 't')


USE_GPU = os.getenv('USE_GPU', "True").lower() not in ('false', 'f', '0', 'off', 'n', 'no')

USE_SMALL_MODELS = _cast_bool_env_var(os.environ.get("SUNO_USE_SMALL_MODELS", "False"))
GLOBAL_ENABLE_MPS = _cast_bool_env_var(os.environ.get("SUNO_ENABLE_MPS", "False"))
OFFLOAD_CPU = _cast_bool_env_var(os.environ.get("SUNO_OFFLOAD_CPU", "False"))



import re
import unicodedata


def encode_path_safe(filename: str, allow_unicode=False):
    """
    Makes a string path safe by removing / replacing not by the os allowed patterns.
    This converts:
    spaces 2 dashes, repeated dashes 2 single dashes, remove non alphanumerics, underscores, or hyphen, string 2 lowercase
    strip
    """
    filename = str(filename)
    if allow_unicode:
        filename = unicodedata.normalize('NFKC', filename)
    else:
        filename = unicodedata.normalize('NFKD', filename).encode('ascii', 'ignore').decode('ascii')
    filename = re.sub(r'[^\w\s-]', '', filename.lower())
    return re.sub(r'[-\s]+', '', filename).strip('-_')


def get_cpu_or_gpu() -> str:
    import torch
    if torch.cuda.is_available():
        return 'cuda'
    else:
        return 'cpu'


def create_progress_tracker(original_update_func: callable, steps: list = None):
    """
    Create a progress tracking function with predefined step weights.

    :param steps: A list of tuples (step_name, step_weight)
                  Weights should sum to 100
    :return: Callable progress update function
    """
    if not steps or not isinstance(steps, list) or not all(
            isinstance(step, tuple) and len(step) == 2 for step in steps
    ):
        steps = [("one", 100)]

    total_steps = sum(step[1] for step in steps)
    if total_steps != 100:
        raise ValueError("Step weights must sum to 100")

    current_progress = 0.0
    prev_progress = 0.0
    step_index = 0

    def progress_update_func_block(x):
        nonlocal current_progress, step_index, prev_progress

        # Calculate actual progress for current step
        current_step_weight = steps[step_index][1]
        step_progress = (current_step_weight / 100) * x
        total_current_progress = current_progress + step_progress

        # Call external progress update if defined
        if original_update_func:
            if prev_progress <= total_current_progress:
                original_update_func(total_current_progress)
                prev_progress = total_current_progress

        # Move to next step if progress reaches 100%
        if x >= 100:
            current_progress += current_step_weight
            step_index = min(step_index + 1, len(steps) - 1)
            prev_progress = 0.0

        return total_current_progress

    return progress_update_func_block
